import math
import numpy as np


class CartPole:
    """
    CartPole environment modelled as an MDP

    ...

    Attributes
    ----------
    gravity: float
        gravitational acceleration
    masscart: float
        mass of the cart
    masspole: float
        mass of the pole
    length: float
        length of the pole
    force_mag: float
        magnitude of the applied force

    Methods
    -------
    reset()
        resets the MDP state
    step(action)
        changes the state of the MDP upon executing an action, where the action set is {right,up,left,down,stay}
    state_label(state)
        outputs the label of input state
    """

    def __init__(self):
        self.gravity = 9.8
        self.masscart = 1.0
        self.masspole = 0.1
        self.total_mass = (self.masspole + self.masscart)
        self.length = 0.5
        self.polemass_length = (self.masspole * self.length)
        self.force_mag = 10.0
        self.tau = 0.02
        self.kinematics_integrator = 'euler'
        self.theta_threshold_radians = 12 * 2 * math.pi / 360
        self.x_threshold = 2.4
        self.current_state = np.random.uniform(low=-0.05, high=0.05, size=(4,))
        self.action_space = [-1, 1]

    def step(self, action):
        x, x_dot, theta, theta_dot = self.current_state
        force = self.force_mag if action[0] > 0 else -self.force_mag
        costheta = math.cos(theta)
        sintheta = math.sin(theta)
        temp = (force + self.polemass_length * theta_dot ** 2 * sintheta) / self.total_mass
        thetaacc = (self.gravity * sintheta - costheta * temp) / (self.length * (4.0 / 3.0 - self.masspole * costheta ** 2 / self.total_mass))
        xacc = temp - self.polemass_length * thetaacc * costheta / self.total_mass

        if self.kinematics_integrator == 'euler':
            x = x + self.tau * x_dot
            x_dot = x_dot + self.tau * xacc
            theta = theta + self.tau * theta_dot
            theta_dot = theta_dot + self.tau * thetaacc
        else:  # semi-implicit euler
            x_dot = x_dot + self.tau * xacc
            x = x + self.tau * x_dot
            theta_dot = theta_dot + self.tau * thetaacc
            theta = theta + self.tau * theta_dot

        self.current_state = np.array((x, x_dot, theta, theta_dot))

        return self.current_state

    def reset(self):
        self.current_state = np.random.uniform(low=-0.05, high=0.05, size=(4,))

    def state_label(self, state):
        x, x_dot, theta, theta_dot = state
        if x < -self.x_threshold or \
                x > self.x_threshold or \
                theta < -self.theta_threshold_radians or \
                theta > self.theta_threshold_radians:
            return 'd'
        else:
            return 'u'



import os
import random
import numpy as np
from matplotlib.image import imread


class MarsRover:
    """
    An MDP whose labels depend on a background image

    ...

    Attributes
    ----------
    image: string
        path to a layout file

    Methods
    -------
    reset()
        resets the MDP state, including ghosts and agent
    step(action)
        changes the state of the MDP upon executing an action, where the action set is {right,up,left,down,stay}
    state_label(state)
        outputs the label of input state
    """

    def __init__(
            self,
            image='marsrover_1.png',
            initial_state=None
    ):
        self.background = imread(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'layouts', image))
        self.width = self.background.shape[0]
        self.height = self.background.shape[1]
        self.labels = self.background.__array__()
        self.initial_state = initial_state
        if self.initial_state is None:
            self.initial_state = np.array([35, 25], dtype=np.float32)
        self.current_state = self.initial_state
        # directional actions
        self.action_space = [
            "right",
            "up",
            "left",
            "down",
            "stay"
        ]

    def reset(self):
        self.current_state = self.initial_state.copy()

    def step(self, action):
        # agent movement dynamics:
        # # stochasticity
        traversed_distance = 10*random.random()
        noise = np.array([random.uniform(-0.1, 0.5), random.uniform(-0.1, 0.5)])
        if action == 'right':
            next_state = self.current_state + [0, traversed_distance] + noise
        elif action == 'up':
            next_state = self.current_state + [-traversed_distance, 0] + noise
        elif action == 'left':
            next_state = self.current_state + [0, -traversed_distance] + noise
        elif action == 'down':
            next_state = self.current_state + [traversed_distance, 0] + noise
        elif action == 'stay':
            next_state = self.current_state + [0, 0] + noise

        # check for boundary violations
        if next_state[0] > self.width-1:
            next_state[0] = self.width-1
        if next_state[1] > self.height-1:
            next_state[1] = self.height-1
        if next_state[0] < 0:
            next_state[0] = 0
        if next_state[1] < 0:
            next_state[1] = 0

        # update current state
        self.current_state = next_state
        return next_state

    def state_label(self, state):
        # note: labels are inevitably discrete when reading an image file
        # thus in the following we look where the continuous state lies within
        # the image rgb matrix
        low_bound = [abs(state[i]-int(state[i])) for i in range(len(state))]
        high_bound = [1-low_bound[i] for i in range(len(state))]
        state_rgb_indx = []
        for i in range(len(state)):
            if low_bound[i] <= high_bound[i]:
                # check for boundary
                if int(state[i]) > self.background.shape[i] - 1:
                    state_rgb_indx.append(self.background.shape[i] - 1)
                else:
                    state_rgb_indx.append(int(state[i]))
            else:
                # check for boundary
                if int(state[i]) + 1 > self.background.shape[i] - 1:
                    state_rgb_indx.append(self.background.shape[i] - 1)
                else:
                    state_rgb_indx.append(int(state[i])+1)

        if list(self.labels[state_rgb_indx[0], state_rgb_indx[1]]) == list(self.labels[0, 199]) or \
                list(self.labels[state_rgb_indx[0], state_rgb_indx[1]]) == list(self.labels[144, 199]):
            return 'unsafe'
        elif list(self.labels[state_rgb_indx[0], state_rgb_indx[1]]) == list(self.labels[45, 152]):
            return 'goal1'
        elif list(self.labels[state_rgb_indx[0], state_rgb_indx[1]]) == list(self.labels[62, 16]):
            return 'goal2'
        else:
            return 'safe'




from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# an example slippery grid
# only the labelling function needs to be specified

# create a SlipperyGrid object
slp_lrg = SlipperyGrid(shape=[40, 40], initial_state=[2, 0], slip_probability=0.05)

# define the labellings
labels = np.empty([slp_lrg.shape[0], slp_lrg.shape[1]], dtype=object)
labels[0:40, 0:40] = 'safe'
labels[20:28, 31:39] = 'goal1'
labels[31:39, 31:39] = 'goal2'
labels[20:28, 20:28] = 'goal3'
labels[0:8, 31:39] = 'goal4'

# override the labels
slp_lrg.labels = labels


from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# create a SlipperyGrid object
minecraft = SlipperyGrid(shape=[10, 10], initial_state=[9, 2], slip_probability=0)

# define the labellings
labels = np.empty([minecraft.shape[0], minecraft.shape[1]], dtype=object)
labels[0:10, 0:10] = 'safe'
labels[0:3, 5] = 'obstacle'
labels[2, 7:10] = 'obstacle'
labels[0][0] = labels[4][5] = labels[8][1] = labels[8][7] = 'grass'
labels[2][2] = labels[7][3] = labels[5][7] = labels[9][9] = 'wood'
labels[0][3] = labels[4][0] = labels[6][8] = labels[9][4] = 'iron'
labels[6][1] = labels[6][5] = labels[4][9] = 'work_bench'
labels[2][4] = labels[9][0] = labels[7][7] = 'tool_shed'
labels[0][7] = 'gold'

# override the labels
minecraft.labels = labels



from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# an example slippery grid "assets/layout1.png" or "https://i.imgur.com/CzSbaYi.png"
# dark blue = safe
# red = unsafe
# cyan = goal1
# yellow = goal2
# only the labelling function needs to be specified

# create a SlipperyGrid object
gridworld_1 = SlipperyGrid(initial_state=[0, 0])

# defines the labelling image
labels = np.empty([gridworld_1.shape[0], gridworld_1.shape[1]], dtype=object)
labels[0:40, 0:40] = 'safe'
labels[25:33, 7:15] = 'unsafe'
labels[7:15, 25:33] = 'unsafe'
labels[15:25, 15:25] = 'goal1'
labels[33:40, 0:7] = 'goal2'

# override the labels
gridworld_1.labels = labels



from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# A modified version of OpenAI Gym FrozenLake
# only the labelling function needs to be specified

# create a SlipperyGrid object
FrozenLake = SlipperyGrid(shape=[12, 10],
                          initial_state=[0, 5],
                          slip_probability=0.1,
                          sink_states=[[4, 7], [5, 8]]
                          )

# define the labellings
labels = np.empty([FrozenLake.shape[0], FrozenLake.shape[1]], dtype=object)
labels[0:12, 0:10] = 'safe'
labels[3:5, 3:5] = 'unsafe'
labels[4:6, 7:9] = 'goal1'
labels[7:9, 7:9] = 'goal2'
labels[6:8, 3:5] = 'goal3'
labels[7:9, 0:2] = 'goal4'

# override the labels
FrozenLake.labels = labels

# FrozenLake doesn't have the action "stay"
FrozenLake.action_space = [
    "right",
    "up",
    "left",
    "down",
]



from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# an example slippery grid "assets/layout1.png" or "https://i.imgur.com/CzSbaYi.png"
# dark blue = safe
# red = unsafe
# cyan = goal1
# yellow = goal2
# only the labelling function needs to be specified

# create a SlipperyGrid object
gridworld_2 = SlipperyGrid(initial_state=[0, 39])

# defines the labelling image
labels = np.empty([gridworld_2.shape[0], gridworld_2.shape[1]], dtype=object)
labels[0:40, 0:40] = 'safe'
labels[25:33, 0:40] = 'unsafe'
labels[25:33, 12:28] = 'safe'
labels[0:8, 0:8] = 'goal1'
labels[33:40, 0:40] = 'goal2'

# override the labels
gridworld_2.labels = labels




from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# A modified version of OpenAI Gym FrozenLake
# only the labelling function needs to be specified

# create a SlipperyGrid object
FrozenLake = SlipperyGrid(shape=[40, 40],
                          initial_state=[0, 20],
                          slip_probability=0.1
                          )

# define the labellings
labels = np.empty([FrozenLake.shape[0], FrozenLake.shape[1]], dtype=object)
labels[0:40, 0:40] = 'safe'
labels[4:8, 9:13] = 'unsafe'
labels[20:28, 31:39] = 'goal1'
labels[31:39, 31:39] = 'goal2'
labels[20:28, 20:28] = 'goal3'
labels[0:4, 15:19] = 'goal4'

# override the labels
FrozenLake.labels = labels

# FrozenLake doesn't have the action "stay"
FrozenLake.action_space = [
    "right",
    "up",
    "left",
    "down",
]


import random
from operator import add


class SlipperyGrid:
    """
    Slippery grid-world modelled as an MDP

    ...

    Attributes
    ----------
    shape: list
        1d list with two elements: 1st element is the num of row cells and the 2nd is the num of column cells (default [40, 40])
    initial_state : list
        1d list with two elements (default [0, 39])
    slip_probability: float
        probability of slipping (default 0.15)
    sink_states : list
        sinks states if any (default [])

    Methods
    -------
    reset()
        resets the MDP state
    step(action)
        changes the state of the MDP upon executing an action, where the action set is {right,up,left,down,stay}
    state_label(state)
        outputs the label of input state
    """

    def __init__(
            self,
            shape=None,
            initial_state=None,
            slip_probability=0.15,
            sink_states=None
    ):
        if sink_states is None:
            sink_states = []
        if shape is None:
            self.shape = [40, 40]
        else:
            self.shape = shape
        if initial_state is None:
            self.initial_state = [0, 39]
        else:
            self.initial_state = initial_state

        self.current_state = self.initial_state.copy()
        self.slip_probability = slip_probability
        self.sink_states = sink_states
        self.labels = None

        # directional actions
        self.action_space = [
            "right",
            "up",
            "left",
            "down",
            "stay"
        ]

    def reset(self):
        self.current_state = self.initial_state.copy()

    def step(self, action):
        # check if the agent is in a sink state
        if self.current_state in self.sink_states:
            next_state = self.current_state
        else:
            # slipperiness
            if random.random() < self.slip_probability:
                action = random.choice(self.action_space)

            # grid movement dynamics:
            if action == 'right':
                next_state = list(map(add, self.current_state, [0, 1]))
            elif action == 'up':
                next_state = list(map(add, self.current_state, [-1, 0]))
            elif action == 'left':
                next_state = list(map(add, self.current_state, [0, -1]))
            elif action == 'down':
                next_state = list(map(add, self.current_state, [1, 0]))
            elif action == 'stay':
                next_state = self.current_state

            # check for boundary violations
            if next_state[0] == self.shape[0]:
                next_state[0] = self.shape[0] - 1
            if next_state[1] == self.shape[1]:
                next_state[1] = self.shape[1] - 1
            if -1 in next_state:
                next_state[next_state.index(-1)] = 0

            # check for obstacles
            if 'obstacle' in self.state_label(next_state):
                next_state = self.current_state

        # update current state
        self.current_state = next_state
        return next_state

    def state_label(self, state):
        return self.labels[state[0], state[1]]



from lcrl.automata.ldba import LDBA

# an example automaton for "wood then tool_shed" or "F (wood & XF (tool_shed))"
# only the automaton "step" function and the "accepting_sets" attribute need to be specified.
# "accepting_sets" is a list of lists for Generalised Büchi Accepting condition (https://bit.ly/ldba_paper)
minecraft_1 = LDBA(accepting_sets=[[2]])


# "step" function for the automaton transitions (input: label, output: automaton_state, un-accepting sink state is "-1")
def step(self, label):
    # state 0
    if self.automaton_state == 0:
        if 'wood' in label:
            self.automaton_state = 1
        else:
            self.automaton_state = 0
    # state 1
    elif self.automaton_state == 1:
        if 'tool_shed' in label:
            self.automaton_state = 2
        else:
            self.automaton_state = 1
    # state 2
    elif self.automaton_state == 2:
        self.automaton_state = 2
    # step function returns the new automaton state
    return self.automaton_state


# now override the step function
LDBA.step = step.__get__(minecraft_1, LDBA)

# finally, does the LDBA contains an epsilon transition? if so then
# for each state with outgoing epsilon-transition define a different epsilon
# example: <LDBA_object_name>.epsilon_transitions = {0: ['epsilon_0'], 4: ['epsilon_1']}
# "0" and "4" are automaton_states
# (for more details refer to https://bit.ly/ldba_paper)



from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# only the labelling function needs to be specified

# create a SlipperyGrid object
robot_surv = SlipperyGrid(shape=[5, 5], initial_state=[1, 0], slip_probability=0.05)

# define the labellings
labels = np.empty([robot_surv.shape[0], robot_surv.shape[1]], dtype=object)
labels[0:5, 0:5] = 'safe'
labels[0, 0] = labels[1, 2] = labels[1, 3] = labels[4, 2] = 'obstacle'
labels[4, 0] = 'goal1'
labels[0, 4] = 'goal2'

# override the labels
robot_surv.labels = labels



from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# OpenAI Gym FrozenLake 4x4
# only the labelling function needs to be specified

# create a SlipperyGrid object
FrozenLake = SlipperyGrid(shape=[4, 4], initial_state=[0, 0], slip_probability=0.05, sink_states=[[3, 3]])

# define the labellings
labels = np.empty([FrozenLake.shape[0], FrozenLake.shape[1]], dtype=object)
labels[0:4, 0:4] = 'safe'
labels[1, 1] = labels[1, 3] = labels[2, 3] = labels[3, 0] = 'unsafe'
labels[0, 3] = 'goal1'
labels[3, 3] = 'goal2'

# override the labels
FrozenLake.labels = labels

# FrozenLake doesn't have the action "stay"
FrozenLake.action_space = [
    "right",
    "up",
    "left",
    "down",
]



import numpy as np


class LDBA:
    """
    Limit Deterministic Büchi Automaton (for more details refer to https://bit.ly/ldba_paper)

    ...

    Attributes
    ----------
    initial_automaton_state : int
        initial automaton state (default 0)
    accepting_sets : list
        the set of accepting sets for Generalised Büchi Accepting condition (more details here https://bit.ly/ldba_paper)

    Methods
    -------
    reset()
        resets the automaton state and its accepting frontier function
    step(label)
        changes and returns the state of the automaton (self.automaton_state) upon reading a label (un-accepting sink state is "-1")
    accepting_frontier_function(automaton_state)
        checks if the automaton state is in the accepting frontier set in order to update it
    """

    def __init__(self, initial_automaton_state=0, accepting_sets=None):
        self.initial_automaton_state = initial_automaton_state
        self.automaton_state = self.initial_automaton_state
        self.accepting_frontier_set = accepting_sets
        self.accepting_sets = accepting_sets
        self.epsilon_transitions = {}

    def reset(self):
        self.automaton_state = self.initial_automaton_state
        self.accepting_frontier_set = self.accepting_sets.copy()

    def step(self, label):
        pass

    def accepting_frontier_function(self, next_automaton_state):
        # for more details refer to the tool paper

        # remove the sets that have intersection with next_automaton_state
        indeces_to_remove = []
        for i in range(len(self.accepting_frontier_set)):
            if next_automaton_state in self.accepting_frontier_set[i]:
                indeces_to_remove.append(i)
        self.accepting_frontier_set = \
            np.delete(self.accepting_frontier_set, indeces_to_remove, axis=0).tolist()

        if indeces_to_remove and not self.accepting_frontier_set:
            self.accepting_frontier_set = self.accepting_sets.copy()
            self.accepting_frontier_set = \
                np.delete(self.accepting_frontier_set, indeces_to_remove, axis=0).tolist()

        # return a positive flag if a set has been removed from the accepting frontier set
        if indeces_to_remove:
            return 1
        # return zero otherwise
        else:
            return 0



import os
import random
from operator import add
import numpy as np


class PacMan:
    """
    Arcade Pacman modelled as an MDP

    ...

    Attributes
    ----------
    layout: string
        path to a layout file

    Methods
    -------
    process_layout()
        creates a labelling function and its associated graph from a '.py' file
        each character in a '.py' file represents a different type of object:
         % - Wall
         . - Token
         o - Food
         G - Ghost
         P - Pacman
        Other characters are ignored
        For built-in examples see `./environments/layouts/`
    reset()
        resets the MDP state, including ghosts and agent
    step(action)
        changes the state of the MDP upon executing an action, where the action set is {right,up,left,down,stay}
    state_label(state)
        outputs the label of input state
    state2cell(state)
        converts the agent / ghost state to its cell number in the grid
    cell2state(int)
        converts a cell number to coordinates of that cell
    """

    def __init__(
            self,
            layout='small'
    ):
        file = open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'layouts', layout) + '.py')
        self.layoutText = [line.strip() for line in file]
        file.close()
        self.width = len(self.layoutText[0])
        self.height = len(self.layoutText)
        self.labels = np.empty([self.width, self.height], dtype=object)
        self.agent_state = []
        self.ghosts_state = []
        self.mdp_graph = {}
        self.agent_initial_state = []
        self.ghosts_initial_state = []
        self.current_state = []
        # directional actions
        self.action_space = [
            "right",
            "up",
            "left",
            "down",
            "stay"
        ]
        self.process_layout()

    def process_layout(self):
        food_counter = 0
        max_y = self.height - 1
        for y in range(self.height):
            for x in range(self.width):
                layout_char = self.layoutText[max_y - y][x]
                if layout_char == '%':
                    self.labels[x][y] = 'obstacle'
                elif layout_char == '.':
                    self.labels[x][y] = 'token'
                elif layout_char == 'o':
                    food_counter += 1
                    self.labels[x][y] = 'food' + str(food_counter)
                elif layout_char == 'P':
                    self.agent_state = self.location2cell([x, y])
                    self.agent_initial_state = self.location2cell([x, y])
                elif layout_char in ['G']:
                    self.ghosts_state.append(self.location2cell([x, y]))
                    self.ghosts_initial_state.append(self.location2cell([x, y]))
        for y in range(1, self.height - 1):
            for x in range(1, self.width - 1):
                graph_vertex = self.location2cell([x, y])
                self.mdp_graph[graph_vertex] = []
                for a in self.action_space:
                    self.agent_state = self.location2cell([x, y])
                    neightbour_cell = self.step(a, False)
                    if neightbour_cell[0] not in self.mdp_graph[graph_vertex]:
                        self.mdp_graph[graph_vertex].append(neightbour_cell[0])
        self.reset()

    def reset(self):
        self.agent_state = self.agent_initial_state
        self.ghosts_state = self.ghosts_initial_state.copy()
        self.current_state = [self.agent_state] + self.ghosts_state

    def step(self, action, ghosts_moving=1):
        # agent movement dynamics:
        if action == 'right':
            agent_next_location = list(map(add, self.cell2location(self.agent_state), [0, 1]))
        elif action == 'up':
            agent_next_location = list(map(add, self.cell2location(self.agent_state), [-1, 0]))
        elif action == 'left':
            agent_next_location = list(map(add, self.cell2location(self.agent_state), [0, -1]))
        elif action == 'down':
            agent_next_location = list(map(add, self.cell2location(self.agent_state), [1, 0]))
        elif action == 'stay':
            agent_next_location = self.cell2location(self.agent_state)

        # check for obstacles
        if self.labels[agent_next_location[0]][agent_next_location[1]] is not None \
                and 'obstacle' in self.labels[agent_next_location[0]][agent_next_location[1]]:
            agent_next_location = self.cell2location(self.agent_state)

        # update agent state
        self.agent_state = self.location2cell(agent_next_location)

        # how ghosts chase the agent
        if ghosts_moving:
            for i in range(len(self.ghosts_state)):
                # with probability of 40% the ghosts chase the pacman
                if random.random() < 0.4:
                    try:
                        self.ghosts_state[i] = \
                            self.find_shortest_path(self.mdp_graph,
                                                    self.ghosts_state[i],
                                                    self.agent_state)[1]
                    except IndexError:
                        self.ghosts_state[i] = \
                            self.find_shortest_path(self.mdp_graph,
                                                    self.ghosts_state[i],
                                                    self.agent_state)[0]
                # with probability of 60% the ghosts take a random action
                else:
                    ghost_i_next_location = list(map(add, self.cell2location(self.ghosts_state[i]),
                                                     random.sample([[1, 0],
                                                                    [-1, 0],
                                                                    [0, -1],
                                                                    [1, 0]], 1)[0]))
                    # check for obstacles
                    if self.labels[ghost_i_next_location[0]][ghost_i_next_location[1]] is not None \
                            and 'obstacle' in self.labels[ghost_i_next_location[0]][ghost_i_next_location[1]]:
                        ghost_i_next_location = self.cell2location(self.ghosts_state[i])

                    # update ghost state
                    self.ghosts_state[i] = self.location2cell(ghost_i_next_location)
        # return the MDP state
        mdp_state = [self.agent_state] + self.ghosts_state
        self.current_state = mdp_state
        return mdp_state

    def state_label(self, state):
        location = self.cell2location(state[0])
        if self.agent_state in self.ghosts_state:
            return 'ghost'
        else:
            return self.labels[location[0]][location[1]]

    def location2cell(self, state):
        return state[1] + state[0] * self.height

    def cell2location(self, cell_num):
        return [int(cell_num / self.height), cell_num % self.height]

    def find_shortest_path(self, graph, start, end, path=None):
        if path is None:
            path = []
        path = path + [start]
        if start == end:
            return path
        if start not in graph.keys():
            return None
        shortest = None
        for node in graph[start]:
            if node not in path:
                newpath = self.find_shortest_path(graph, node, end, path)
                if newpath:
                    if not shortest or len(newpath) < len(shortest):
                        shortest = newpath
        return shortest



import numpy as np
import random


class LCRL:
    """
    LCRL

    ...

    Attributes
    ----------
    MDP : an object with the following properties
        (1) a "step(action)" function describing the dynamics
        (2) a "reset()" function that resets the state to an initial state
        (3) a "state_label(state)" function that maps states to labels
        (4) current state of the MDP is "current_state"
        (5) action space is "action_space" and all actions are enabled in each state
    LDBA : an object of ./src/automata/LDBA.py
        a limit-deterministic Büchi automaton
    discount_factor: float
        discount factor (default 0.9)
    learning_rate: float
        learning rate (default 0.9)
    epsilon: float
        tuning parameter for the epsilon-greedy exploration scheme (default 0.1)

    Methods
    -------
    train_ql(number_of_episodes, iteration_threshold, Q_initial_value)
        employs Episodic Q-learning to synthesise an optimal policy over the product MDP
    train_nfq(number_of_episodes, iteration_threshold, nfq_replay_buffer_size, num_of_hidden_neurons)
        employs Episodic Neural Fitted Q-Iteration to synthesise an optimal policy over the product MDP
    train_ddpg(number_of_episodes, iteration_threshold, ddpg_replay_buffer_size, num_of_hidden_neurons)
        employs Episodic Deep Deterministic Policy Gradient to synthesise an optimal policy over the product MDP
    reward(automaton_state)
        shapes the reward function according to the automaton accepting frontier set
        for more details refer to the tool paper
    action_space_augmentation()
        augments the action space whenever an epsilon transition is expected
        for more details refer to the tool paper
    """

    def __init__(
            self, MDP=None,
            LDBA=None,
            discount_factor=0.9,
            learning_rate=0.9,
            decaying_learning_rate=True,
            epsilon=0.15
    ):
        if MDP is None:
            raise Exception("LCRL expects an MDP object as input")
        self.MDP = MDP
        if LDBA is None:
            raise Exception("LCRL expects an LDBA object as input")
        self.LDBA = LDBA
        self.epsilon_transitions_exists = True if LDBA.epsilon_transitions.__len__() > 0 else False
        self.gamma = discount_factor
        self.alpha = learning_rate
        self.decay_lr = decaying_learning_rate
        if self.decay_lr:
            self.alpha_initial_value = learning_rate
            self.alpha_final_value = 0.1
        self.epsilon = epsilon
        self.path_length = []
        self.Q = {}
        self.replay_buffers = {}
        self.Q_initial_value = 0
        self.early_interruption = 0
        self.q_at_initial_state = []
        self.successes_in_test = 0
        # ##### testing area ##### #
        self.test = False
        # ######################## #

    def train_ql(
            self, number_of_episodes,
            iteration_threshold,
            Q_initial_value=0
    ):
        self.MDP.reset()
        self.LDBA.reset()
        self.Q_initial_value = Q_initial_value

        if self.LDBA.accepting_sets is None:
            raise Exception('LDBA object is not defined properly. Please specify the "accepting_set".')

        # product MDP: synchronise the MDP state with the automaton state
        current_state = self.MDP.current_state + [self.LDBA.automaton_state]
        product_MDP_action_space = self.MDP.action_space
        epsilon_transition_taken = False

        # check for epsilon-transitions at the current automaton state
        if self.epsilon_transitions_exists:
            product_MDP_action_space = self.action_space_augmentation()

        # initialise Q-value outside the main loop
        self.Q[str(current_state)] = {}
        for action_index in range(len(product_MDP_action_space)):
            self.Q[str(current_state)][product_MDP_action_space[action_index]] = Q_initial_value

        # main loop
        try:
            episode = 0
            self.path_length = [float("inf")]
            while episode < number_of_episodes:
                episode += 1
                self.MDP.reset()
                self.LDBA.reset()
                current_state = self.MDP.current_state + [self.LDBA.automaton_state]

                # check for epsilon-transitions at the current automaton state
                if self.epsilon_transitions_exists:
                    product_MDP_action_space = self.action_space_augmentation()

                # Q value at the initial state
                Q_at_initial_state = []
                for action_index in range(len(product_MDP_action_space)):
                    Q_at_initial_state.append(self.Q[str(current_state)][product_MDP_action_space[action_index]])
                # value function at the initial state
                self.q_at_initial_state.append(max(Q_at_initial_state))
                print('episode:' + str(episode)
                      + ', value function at s_0=' + str(max(Q_at_initial_state))
                      # + ', trace length=' + str(self.path_length[len(self.path_length) - 1])
                      # + ', learning rate=' + str(self.alpha)
                      # + ', current state=' + str(self.MDP.current_state)
                      )
                iteration = 0
                path = current_state

                # annealing the learning rate
                if self.decay_lr:
                    self.alpha = max(self.alpha_final_value,
                                     ((self.alpha_final_value - self.alpha_initial_value) / (0.8 * number_of_episodes))
                                     * episode + self.alpha_initial_value)

                # each episode loop
                while self.LDBA.accepting_frontier_set and \
                        iteration < iteration_threshold and \
                        self.LDBA.automaton_state != -1:
                    iteration += 1

                    # find the action with max Q at the current state
                    Qs = []
                    for action_index in range(len(product_MDP_action_space)):
                        Qs.append(self.Q[str(current_state)][product_MDP_action_space[action_index]])
                    maxQ_action_index = random.choice(np.where(Qs == np.max(Qs))[0])
                    maxQ_action = product_MDP_action_space[maxQ_action_index]

                    # check if an epsilon-transition is taken
                    if self.epsilon_transitions_exists and \
                            maxQ_action_index > len(self.MDP.action_space) - 1:
                        epsilon_transition_taken = True

                    # product MDP modification (for more details refer to the tool paper)
                    if epsilon_transition_taken:
                        next_MDP_state = self.MDP.current_state
                        next_automaton_state = self.LDBA.step(maxQ_action)
                    else:
                        # epsilon-greedy policy
                        if random.random() < self.epsilon:
                            next_MDP_state = self.MDP.step(random.choice(self.MDP.action_space))
                        else:
                            next_MDP_state = self.MDP.step(maxQ_action)
                        next_automaton_state = self.LDBA.step(self.MDP.state_label(next_MDP_state))

                    # product MDP: synchronise the automaton with MDP
                    next_state = next_MDP_state + [next_automaton_state]

                    # check for epsilon-transitions at the next automaton state
                    if self.epsilon_transitions_exists:
                        product_MDP_action_space = self.action_space_augmentation()

                    # Q values of the next state
                    Qs_prime = []
                    if str(next_state) not in self.Q.keys():
                        self.Q[str(next_state)] = {}
                        for action_index in range(len(product_MDP_action_space)):
                            self.Q[str(next_state)][product_MDP_action_space[action_index]] = Q_initial_value
                            Qs_prime.append(Q_initial_value)
                    else:
                        for action_index in range(len(product_MDP_action_space)):
                            Qs_prime.append(self.Q[str(next_state)][product_MDP_action_space[action_index]])

                    # update the accepting frontier set
                    if not epsilon_transition_taken:
                        reward_flag = self.LDBA.accepting_frontier_function(next_automaton_state)
                    else:
                        reward_flag = 0
                        epsilon_transition_taken = False

                    if reward_flag > 0:
                        state_dep_gamma = self.gamma
                    else:
                        state_dep_gamma = 1

                    # Q update
                    self.Q[str(current_state)][maxQ_action] = \
                        (1 - self.alpha) * self.Q[str(current_state)][maxQ_action] + \
                        self.alpha * (self.reward(reward_flag) + state_dep_gamma * np.max(Qs_prime))

                    if self.test:
                        print(str(maxQ_action)
                              + ' | ' + str(next_state)
                              + ' | ' + self.MDP.state_label(next_MDP_state)
                              + ' | ' + str(reward_flag)
                              + ' | ' + str(self.Q[str(current_state)][maxQ_action]))

                    # update the state
                    current_state = next_state.copy()
                    path.append(current_state)

                # append the path length
                self.path_length.append(len(path))

        except KeyboardInterrupt:
            print('\nTraining exited early.')
            try:
                is_save = input(
                    'Would you like to save the training data? '
                    'If so, type in "y", otherwise, interrupt with CTRL+C. ')
            except KeyboardInterrupt:
                print('\nExiting...')

            if is_save == 'y' or is_save == 'Y':
                print('Saving...')
                self.early_interruption = 1

    def train_nfq(
            self, number_of_episodes,
            iteration_threshold,
            nfq_replay_buffer_size,
            num_of_hidden_neurons=256
    ):
        import tensorflow as tf
        from tensorflow import keras
        tf.get_logger().setLevel('ERROR')
        self.MDP.reset()
        self.LDBA.reset()

        if self.LDBA.accepting_sets is None:
            raise Exception('LDBA object is not defined properly. Please specify the "accepting_set". ')

        # ## exploration phase ## #
        # nfq_modules
        self.Q = {}
        self.replay_buffers = {}
        state_dimension = len(self.MDP.current_state)
        action_dimension = 1
        nfq_input_dim = state_dimension + action_dimension
        product_MDP_action_space = self.MDP.action_space
        epsilon_transition_taken = 0

        # initiate an NFQ module
        model_0 = keras.Sequential([
            keras.layers.Dense(num_of_hidden_neurons, input_dim=nfq_input_dim, activation=tf.nn.relu),
            keras.layers.Dense(1, activation=tf.nn.leaky_relu)])
        model_0.compile(loss='mean_squared_error',
                        metrics=['mean_squared_error'],
                        optimizer='Adam')
        self.Q[self.LDBA.automaton_state] = model_0
        self.replay_buffers[self.LDBA.automaton_state] = []

        # check for epsilon-transitions at the current automaton state
        if self.epsilon_transitions_exists:
            product_MDP_action_space = self.action_space_augmentation()

        try:
            # exploration loop
            episode = 0
            print("sampling...")
            rewarding_paths = []
            while episode < number_of_episodes:
                episode_path = []
                episode += 1
                self.MDP.reset()
                self.LDBA.reset()
                current_state = self.MDP.current_state.tolist() + [self.LDBA.automaton_state]

                # check for epsilon-transitions at the current automaton state
                if self.epsilon_transitions_exists:
                    product_MDP_action_space = self.action_space_augmentation()

                iteration = 0

                if self.decay_lr:
                    try:
                        input(
                            'Decaying learning rate has to be set to False in "nfq". '
                            'Would you like me to set it to False and continue? '
                            'If so, type in "y", otherwise, interrupt with CTRL+C. ')
                    except KeyboardInterrupt:
                        print('\nExiting...')

                # each episode loop
                while self.LDBA.accepting_frontier_set and \
                        iteration < iteration_threshold and \
                        self.LDBA.automaton_state != -1:
                    iteration += 1

                    # find the action with max Q at the current state
                    active_model = self.LDBA.automaton_state

                    action_index = random.randint(0, len(product_MDP_action_space) - 1)
                    maxQ_action = product_MDP_action_space[action_index]

                    # check if an epsilon-transition is taken
                    if self.epsilon_transitions_exists and \
                            action_index > len(self.MDP.action_space) - 1:
                        epsilon_transition_taken = 1

                    # product MDP modification (for more details refer to the tool paper)
                    if epsilon_transition_taken:
                        next_MDP_state = self.MDP.current_state.tolist()
                        next_automaton_state = self.LDBA.step(maxQ_action)
                    else:
                        next_MDP_state = self.MDP.step(maxQ_action).tolist()
                        next_automaton_state = self.LDBA.step(self.MDP.state_label(next_MDP_state))

                    # product MDP: synchronise the automaton with MDP
                    next_state = next_MDP_state + [next_automaton_state]

                    if next_automaton_state == -1:
                        break

                    if self.test:
                        print(str(maxQ_action) + ' | ' + str(next_state) + ' | ' + self.MDP.state_label(next_MDP_state))

                    # check for epsilon-transitions at the next automaton state
                    if self.epsilon_transitions_exists:
                        product_MDP_action_space = self.action_space_augmentation()

                    # update the accepting frontier set
                    if not epsilon_transition_taken:
                        reward_flag = self.LDBA.accepting_frontier_function(next_automaton_state)
                        reward_flag = reward_flag + (np.cos(np.pi*reward_flag)) * 0.1 * random.random()  # to break symmetry
                    else:
                        reward_flag = 0 + 0.1 * random.random()  # to break symmetry
                        epsilon_transition_taken = 0

                    if reward_flag > 0.5:
                        state_dep_gamma = self.gamma
                        # print('reward reached!')
                    else:
                        state_dep_gamma = self.gamma

                    # create an NFQ module if needed
                    if next_automaton_state not in self.Q.keys() and self.LDBA.accepting_frontier_set:
                        # initiate an NFQ module
                        self.Q[next_automaton_state] = \
                            keras.Sequential([
                                keras.layers.Dense(num_of_hidden_neurons, input_dim=nfq_input_dim,
                                                   activation=tf.nn.relu),
                                keras.layers.Dense(1, activation=tf.nn.leaky_relu)])
                        self.Q[next_automaton_state].compile(loss='mean_squared_error',
                                                             metrics=['mean_squared_error'],
                                                             optimizer='Adam')
                        self.replay_buffers[next_automaton_state] = []

                    # save the experience
                    # (state, action, state, reward, state_dependent_gamma, epsilon_transition_taken)
                    # to the replay buffer
                    self.replay_buffers[active_model].append(
                        current_state[0: -1] +
                        [action_index] +
                        next_state[0: -1] +
                        [reward_flag] +
                        [state_dep_gamma] +
                        [epsilon_transition_taken] +
                        [current_state[-1]]
                    )
                    episode_path.append(self.replay_buffers[active_model][-1])
                    if reward_flag > 0.5:
                        rewarding_paths.extend(episode_path)

                    # update the state
                    current_state = next_state

            # exploitation loop
            exp_size = nfq_replay_buffer_size
            high_reward_total = np.array(rewarding_paths)
            sars = {}
            high_reward_sars = {}
            for model_key in self.Q.keys():
                high_reward_sars[model_key] = high_reward_total[high_reward_total[:, -1] == model_key]
                if len(high_reward_sars[model_key]) < exp_size:
                    sars[model_key] = np.array(self.replay_buffers[model_key])
                    uniform_sampled_sars = sars[model_key][
                        np.random.choice(sars[model_key].shape[0], exp_size - len(high_reward_sars[model_key]),
                                         replace=False)]
                    sars[model_key] = np.vstack((high_reward_sars[model_key], uniform_sampled_sars))
                    np.random.shuffle(sars[model_key])
                else:
                    sars[model_key] = high_reward_sars[model_key]

            for model_key in self.Q.keys():
                self.Q[model_key].fit(
                    sars[model_key][:, 0:state_dimension + 1],  # stat - action
                    sars[model_key][:, nfq_input_dim + state_dimension],  # reward
                    epochs=3, verbose=0)

            episode = 0
            self.MDP.reset()
            self.LDBA.reset()
            init_state = self.MDP.current_state.tolist().copy()
            while episode < number_of_episodes:
                # Q value at the initial state
                Q_at_initial_state = []
                for action_index in range(len(product_MDP_action_space)):
                    Q_at_initial_state.append(
                        list(self.Q[0].predict(np.array(init_state + [action_index]).reshape(1, nfq_input_dim))[0])[0]
                    )

                self.q_at_initial_state.append(max(Q_at_initial_state))
                print('episode:' + str(episode)
                      + ', value function at s_0=' + str(self.q_at_initial_state[episode])
                      # + ', trace length=' + str(self.path_length[len(self.path_length) - 1])
                      # + ', learning rate=' + str(self.alpha)
                      # + ', current state=' + str(self.MDP.current_state)
                      )

                for model_key in self.Q.keys():
                    target = np.zeros(len(sars[model_key]))
                    for k in range(len(sars[model_key])):
                        neigh = []
                        if sars[model_key][k, -1] == 1:  # meaning that an epsilon transition was active
                            action_space = list(range(len(self.MDP.action_space)))
                            action_space = action_space + [action_space[-1] + 1]  # for the epsilon transition
                        else:
                            action_space = self.MDP.action_space
                        neigh_input = []
                        for a in range(len(action_space)):
                            neigh_input = np.append(sars[model_key][k, nfq_input_dim:nfq_input_dim + state_dimension],
                                                    a).reshape(1, nfq_input_dim)
                            neigh.append(self.Q[model_key].predict(neigh_input))
                        # target = reward + gamma * max Q(s',a')
                        target[k] = sars[model_key][k, nfq_input_dim + state_dimension] + \
                                    sars[model_key][k, nfq_input_dim + state_dimension + 1] * max(neigh)
                        # terminal state check
                        if sars[model_key][k, nfq_input_dim + state_dimension] > 0.5:
                            target[k] = sars[model_key][k, nfq_input_dim + state_dimension]
                    self.Q[model_key].fit(
                        sars[model_key][:, 0:state_dimension + 1],
                        target.reshape(len(sars[model_key]), 1),
                        epochs=3,
                        verbose=0)

                episode += 1

        except KeyboardInterrupt:
            print('\nTraining exited early.')
            try:
                is_save = input(
                    'Would you like to save the training data? '
                    'If so, type in "y", otherwise, interrupt with CTRL+C. ')
            except KeyboardInterrupt:
                print('\nExiting...')

            if is_save == 'y' or is_save == 'Y':
                print('Saving...')
                self.early_interruption = 1

    def train_ddpg(
            self, number_of_episodes,
            iteration_threshold,
            ddpg_replay_buffer_size,
            num_of_hidden_neurons=32
    ):
        import tensorflow as tf
        tf.get_logger().setLevel('ERROR')
        from keras import layers
        state_dimension = len(self.MDP.current_state)
        action_dimension = 1
        lower_bound = -1.0
        upper_bound = 1.0

        if self.LDBA.accepting_sets is None:
            raise Exception('LDBA object is not defined properly. Please specify the "accepting_set". ')

        class OUActionNoise:
            def __init__(self, mean, std_deviation, theta=0.15, dt=1e-2, x_initial=None):
                self.theta = theta
                self.mean = mean
                self.std_dev = std_deviation
                self.dt = dt
                self.x_initial = x_initial
                self.reset()

            def __call__(self):
                x = (
                    self.x_prev
                    + self.theta * (self.mean - self.x_prev) * self.dt
                    + self.std_dev * np.sqrt(self.dt) * np.random.normal(size=self.mean.shape)
                )
                self.x_prev = x
                return x

            def reset(self):
                if self.x_initial is not None:
                    self.x_prev = self.x_initial
                else:
                    self.x_prev = np.zeros_like(self.mean)

        class Buffer:
            def __init__(self, buffer_capacity=ddpg_replay_buffer_size, batch_size=64, gamma=self.gamma):
                # maximum number of experience samples
                self.buffer_capacity = buffer_capacity
                self.batch_size = batch_size
                self.gamma = gamma

                # num of times record() was called.
                self.buffer_counter = 0

                self.state_buffer = np.zeros((self.buffer_capacity, state_dimension))
                self.action_buffer = np.zeros((self.buffer_capacity, action_dimension))
                self.reward_buffer = np.zeros((self.buffer_capacity, 1))
                self.next_state_buffer = np.zeros((self.buffer_capacity, state_dimension))

            # (s,a,r,s') tuple as input
            def record(self, obs_tuple):
                # buffer capacity index
                index = self.buffer_counter % self.buffer_capacity

                self.state_buffer[index] = obs_tuple[0]
                self.action_buffer[index] = obs_tuple[1]
                self.reward_buffer[index] = obs_tuple[2]
                self.next_state_buffer[index] = obs_tuple[3]

                self.buffer_counter += 1

            @tf.function
            def update(
                    self, state_batch, action_batch, reward_batch, next_state_batch,
            ):
                # training actor and critic
                with tf.GradientTape() as tape:
                    target_actions = target_actor_dict[active_model](next_state_batch, training=True)
                    y = reward_batch + self.gamma * target_critic_dict[active_model](
                        [next_state_batch, target_actions], training=True
                    )
                    critic_value = critic_dict[active_model]([state_batch, action_batch], training=True)
                    critic_loss = tf.math.reduce_mean(tf.math.square(y - critic_value))

                critic_grad = tape.gradient(critic_loss, critic_dict[active_model].trainable_variables)
                critic_optimizer.apply_gradients(
                    zip(critic_grad, critic_dict[active_model].trainable_variables)
                )

                with tf.GradientTape() as tape:
                    actions = actor_dict[active_model](state_batch, training=True)
                    critic_value = critic_dict[active_model]([state_batch, actions], training=True)
                    actor_loss = -tf.math.reduce_mean(critic_value)

                actor_grad = tape.gradient(actor_loss, actor_dict[active_model].trainable_variables)
                actor_optimizer.apply_gradients(
                    zip(actor_grad, actor_dict[active_model].trainable_variables)
                )

            # compute the loss and update parameters
            def learn(self):
                # sampling
                record_range = min(self.buffer_counter, self.buffer_capacity)
                batch_indices = np.random.choice(record_range, self.batch_size)

                state_batch = tf.convert_to_tensor(self.state_buffer[batch_indices])
                action_batch = tf.convert_to_tensor(self.action_buffer[batch_indices])
                reward_batch = tf.convert_to_tensor(self.reward_buffer[batch_indices])
                reward_batch = tf.cast(reward_batch, dtype=tf.float32)
                next_state_batch = tf.convert_to_tensor(self.next_state_buffer[batch_indices])

                self.update(state_batch, action_batch, reward_batch, next_state_batch)

        # soft update
        @tf.function
        def update_target(target_weights, weights, tau):
            for (a, b) in zip(target_weights, weights):
                a.assign(b * tau + a * (1 - tau))

        def get_actor():
            last_init = tf.random_uniform_initializer(minval=-0.003, maxval=0.003)

            inputs = layers.Input(shape=(state_dimension,))
            out = layers.Dense(num_of_hidden_neurons*8, activation="relu")(inputs)
            out = layers.Dense(num_of_hidden_neurons*8, activation="relu")(out)
            outputs = layers.Dense(1, activation="tanh", kernel_initializer=last_init)(out)

            model = tf.keras.Model(inputs, outputs)
            return model

        def get_critic():
            # state as input
            state_input = layers.Input(shape=(state_dimension))
            state_out = layers.Dense(int(num_of_hidden_neurons/2), activation="relu")(state_input)
            state_out = layers.Dense(num_of_hidden_neurons, activation="relu")(state_out)

            # action as input
            action_input = layers.Input(shape=(action_dimension))
            action_out = layers.Dense(num_of_hidden_neurons, activation="relu")(action_input)

            # concatenating
            concat = layers.Concatenate()([state_out, action_out])

            out = layers.Dense(num_of_hidden_neurons*8, activation="relu")(concat)
            out = layers.Dense(num_of_hidden_neurons*8, activation="relu")(out)
            outputs = layers.Dense(1)(out)

            # output
            model = tf.keras.Model([state_input, action_input], outputs)

            return model

        def policy(state, noise_object):
            sampled_actions = tf.squeeze(actor_dict[active_model](state))
            noise = noise_object()
            # Adding noise to action
            sampled_actions = sampled_actions.numpy() + noise

            # We make sure action is within bounds
            legal_action = np.clip(sampled_actions, lower_bound, upper_bound)

            return [np.squeeze(legal_action)]

        exploration_parameter = self.epsilon*3
        ou_noise = OUActionNoise(mean=np.zeros(1), std_deviation=float(exploration_parameter) * np.ones(1))

        # initiate a DDPG module
        actor_model_0 = get_actor()
        actor_dict = {0: actor_model_0}
        critic_model_0 = get_critic()
        critic_dict = {0: critic_model_0}

        target_actor_0 = get_actor()
        target_actor_dict = {0: target_actor_0}
        target_critic_0 = get_critic()
        target_critic_dict = {0: target_critic_0}

        target_actor_dict[0].set_weights(actor_dict[0].get_weights())
        target_critic_dict[0].set_weights(critic_dict[0].get_weights())

        # learning rates
        critic_lr = self.alpha*0.04
        actor_lr = self.alpha*0.02

        critic_optimizer = tf.keras.optimizers.Adam(critic_lr)
        actor_optimizer = tf.keras.optimizers.Adam(actor_lr)

        # soft update parameter
        tau = 0.005

        buffer_0 = Buffer(ddpg_replay_buffer_size, 128)
        buffer_dict = {0: buffer_0}

        epsilon_transition_taken = 0
        ep_reward_list = []

        # training loop
        try:
            for episode in range(number_of_episodes):

                self.MDP.reset()
                self.LDBA.reset()
                prev_state = self.MDP.current_state
                episodic_reward = 0

                # check for epsilon-transitions at the current automaton state
                if self.epsilon_transitions_exists:
                    product_MDP_action_space = self.action_space_augmentation()

                if self.decay_lr:
                    try:
                        input(
                            'Decaying learning rate has to be set to False in "ddpg". '
                            'Would you like me to set it to False and continue? '
                            'If so, type in "y", otherwise, interrupt with CTRL+C. ')
                    except KeyboardInterrupt:
                        print('\nExiting...')

                # each episode loop
                while True:

                    # get an action from the actor
                    active_model = self.LDBA.automaton_state

                    tf_prev_state = tf.expand_dims(tf.convert_to_tensor(prev_state), 0)

                    action = policy(tf_prev_state, ou_noise)

                    if self.epsilon_transitions_exists and \
                            self.LDBA.automaton_state in self.LDBA.epsilon_transitions.keys() and \
                            random.random() > 0.5:
                        epsilon_action = random.choice(product_MDP_action_space[2:])
                        action = [np.squeeze(
                            int(epsilon_action[-1]) + upper_bound
                        )]
                        epsilon_transition_taken = 1

                    # product MDP modification (for more details refer to the tool paper)
                    if epsilon_transition_taken:
                        next_automaton_state = self.LDBA.step(epsilon_action)
                        state = prev_state
                    else:
                        state = self.MDP.step(action)
                        next_automaton_state = self.LDBA.step(self.MDP.state_label(state))

                    # product MDP: synchronise the automaton with MDP
                    next_MDP_state = self.MDP.current_state.tolist()
                    next_state = next_MDP_state + [next_automaton_state]

                    if self.test:
                        print(str(action) + ' | ' + str(next_state) + ' | ' + self.MDP.state_label(next_MDP_state))

                    # check for epsilon-transitions at the next automaton state
                    if self.epsilon_transitions_exists:
                        product_MDP_action_space = self.action_space_augmentation()

                    # update the accepting frontier set
                    if not epsilon_transition_taken:
                        reward_flag = self.LDBA.accepting_frontier_function(next_automaton_state)
                    else:
                        reward_flag = 0
                        epsilon_transition_taken = 0

                    if reward_flag > 0.5:
                        state_dep_gamma = self.gamma
                    else:
                        state_dep_gamma = 1

                    # create a DDPG module if needed
                    if next_automaton_state not in actor_dict.keys() and self.LDBA.accepting_frontier_set:
                        # initiate a DDPG module
                        actor_dict[next_automaton_state] = get_actor()
                        critic_dict[next_automaton_state] = get_critic()
                        target_actor_dict[next_automaton_state] = get_actor()
                        target_critic_dict[next_automaton_state] = get_critic()

                        buffer_dict[next_automaton_state] = Buffer(ddpg_replay_buffer_size, 128)

                        # Making the weights equal initially
                        target_actor_dict[next_automaton_state].set_weights(
                            actor_dict[next_automaton_state].get_weights())
                        target_critic_dict[next_automaton_state].set_weights(
                            critic_dict[next_automaton_state].get_weights())

                    reward = reward_flag - (np.cos(
                        (np.pi / 2) * reward_flag) * 0.3 + 0.02 * random.random())  # to break symmetry
                    buffer_dict[active_model].record((prev_state, action, reward, state))
                    episodic_reward += reward

                    buffer_dict[active_model].learn()
                    update_target(target_actor_dict[active_model].variables, actor_dict[active_model].variables, tau)
                    update_target(target_critic_dict[active_model].variables, critic_dict[active_model].variables, tau)

                    if next_automaton_state == -1:
                        break
                    if reward_flag > 0.5:
                        # print('reward reached!')
                        self.LDBA.reset()

                    prev_state = state

                ep_reward_list.append(episodic_reward)

                avg_reward = np.mean(ep_reward_list[-40:])
                print("episode: {}, average episode reward={}".format(episode, avg_reward))
                self.q_at_initial_state.append(avg_reward)

            self.Q = actor_dict

        except KeyboardInterrupt:
            print('\nTraining exited early.')
            try:
                is_save = input(
                    'Would you like to save the training data? '
                    'If so, type in "y", otherwise, interrupt with CTRL+C. ')
            except KeyboardInterrupt:
                print('\nExiting...')

            if is_save == 'y' or is_save == 'Y':
                print('Saving...')
                self.Q = actor_dict
                self.early_interruption = 1

    def reward(self, reward_flag):
        if reward_flag > 0:
            return 1
        else:
            return 0

    def action_space_augmentation(self):
        if self.LDBA.automaton_state in self.LDBA.epsilon_transitions.keys():
            product_MDP_action_space = self.MDP.action_space + \
                                       self.LDBA.epsilon_transitions[self.LDBA.automaton_state]
        else:
            product_MDP_action_space = self.MDP.action_space
        return product_MDP_action_space



import os
import matplotlib.pyplot as plt
import numpy as np
import tqdm
import imageio.v2 as imageio
from lcrl.environments.SlipperyGrid import SlipperyGrid


def animate(mdp, executed_policy, dir_to_save, labels_value, cmap, norm, patches):
    if isinstance(mdp, SlipperyGrid):
        animation_dir = os.path.join(dir_to_save, 'animation')
        if not os.path.exists(animation_dir):
            os.mkdir(animation_dir)
        print('---------------------------------\n')
        print('Creating a gif for the trained policy:')
        for i in tqdm.tqdm(range(len(executed_policy))):
            if i == 0:
                plt.imshow(labels_value, interpolation='nearest', cmap=cmap, norm=norm)
                path_x, path_y = np.array(executed_policy[0]).T
                plt.legend(handles=patches, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
                plt.scatter(path_y, path_x, c='red', edgecolors='darkred')
                plt.annotate('s_0', (path_y, path_x), fontsize=15, xytext=(20, 20), textcoords="offset points",
                             va="center", ha="left",
                             bbox=dict(boxstyle="round", fc="w"),
                             arrowprops=dict(arrowstyle="->"))
            else:
                plt.imshow(labels_value, interpolation='nearest', cmap=cmap, norm=norm)
                path_x, path_y = np.array(executed_policy[0:i]).T
                inits_y, initis_x = path_y[0], path_x[0]
                plt.scatter(path_y, path_x, c='lime', edgecolors='teal')
                path_x, path_y = np.array(executed_policy[i]).T
                plt.legend(handles=patches, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
                plt.scatter(path_y, path_x, c='red', edgecolors='darkred')
                plt.annotate('s_0', (inits_y, initis_x), fontsize=15, xytext=(20, 20), textcoords="offset points",
                             va="center", ha="left",
                             bbox=dict(boxstyle="round", fc="w"),
                             arrowprops=dict(arrowstyle="->"))
            plt.title('This policy is synthesised by the trained agent')
            plt.savefig(os.path.join(animation_dir, 'image_file_' + str(i) + '.png'))
        images = []
        for file_name in os.listdir(animation_dir):
            if file_name.endswith('.png'):
                file_path = os.path.join(animation_dir, file_name)
                images.append(imageio.imread(file_path))
        imageio.mimsave(os.path.join(animation_dir, 'executed_policy.gif'), images, fps=55)
    else:
        raise NotImplementedError('The animator does not support this environment yet.')



from lcrl.automata.ldba import LDBA

# an example automaton for "goal1 or goal2 while avoiding unsafe" or "(FG goal1 | FG goal2) & G !unsafe"
# automaton image is available in "./assets" or "https://i.imgur.com/gyDED4O.png"
# only the automaton "step" function and the "accepting_sets" attribute need to be specified.
# "accepting_sets" is a list of lists for Generalised Büchi Accepting condition (https://bit.ly/ldba_paper)
goal1_or_goal2 = LDBA(accepting_sets=[[1, 2]])


# "step" function for the automaton transitions (input: label, output: automaton_state, un-accepting sink state is "-1")
def step(self, label):
    # state 0
    if self.automaton_state == 0:
        if 'epsilon_1' in label:
            self.automaton_state = 1
        elif 'epsilon_2' in label:
            self.automaton_state = 2
        elif 'unsafe' in label:
            self.automaton_state = -1  # un-accepting sink state
        else:
            self.automaton_state = 0
    # state 1
    elif self.automaton_state == 1:
        if 'goal1' in label and 'unsafe' not in label:
            self.automaton_state = 1
        else:
            self.automaton_state = -1  # un-accepting sink state
    # state 2
    elif self.automaton_state == 2:
        if 'goal2' in label and 'unsafe' not in label:
            self.automaton_state = 2
        else:
            self.automaton_state = -1  # un-accepting sink state
    # state -1
    elif self.automaton_state == -1:
        self.automaton_state = -1  # un-accepting sink state
    # step function returns the new automaton state
    return self.automaton_state


# now override the step function
LDBA.step = step.__get__(goal1_or_goal2, LDBA)

# finally, does the LDBA contains an epsilon transition? if so then
# for each state with outgoing epsilon-transition define a different epsilon
# example: <LDBA_object_name>.epsilon_transitions = {0: ['epsilon_0'], 4: ['epsilon_1']}
# "0" and "4" are automaton_states
# (for more details refer to https://bit.ly/ldba_paper)
goal1_or_goal2.epsilon_transitions = {0: ['epsilon_0', 'epsilon_1']}



from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# an example slippery grid
# only the labelling function needs to be specified

# create a SlipperyGrid object
slp_sml = SlipperyGrid(shape=[12, 10], initial_state=[2, 0], slip_probability=0.05)

# define the labellings
labels = np.empty([slp_sml.shape[0], slp_sml.shape[1]], dtype=object)
labels[0:12, 0:10] = 'safe'
labels[5][8] = labels[5][9] = labels[6][8] = labels[6][9] = 'goal1'
labels[8][8] = labels[8][9] = labels[9][8] = labels[9][9] = 'goal2'
labels[7][4] = labels[7][5] = labels[8][4] = labels[8][5] = 'goal3'
labels[8][0] = labels[8][1] = labels[9][0] = labels[9][1] = 'goal4'

# override the labels
slp_sml.labels = labels



from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# an example slippery grid "assets/layout3.png" or "https://i.imgur.com/CzSbaYi.png"
# dark blue = safe
# red = unsafe
# cyan = goal1
# yellow = goal2
# only the labelling function needs to be specified

# create a SlipperyGrid object
gridworld_3 = SlipperyGrid(shape=[5, 5], initial_state=[2, 0], slip_probability=0.05)

# defines the labelling image
labels = np.empty([gridworld_3.shape[0], gridworld_3.shape[1]], dtype=object)
labels[0:5, 0:5] = 'safe'
labels[1:4, 2] = 'obstacle'
labels[0:2, 3:5] = 'goal1'
labels[3:5, 3:5] = 'goal2'

# override the labels
gridworld_3.labels = labels



from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# an example slippery grid
# only the labelling function needs to be specified

# create a SlipperyGrid object
slp_med = SlipperyGrid(shape=[20, 20], initial_state=[2, 0], slip_probability=0.05)

# define the labellings
labels = np.empty([slp_med.shape[0], slp_med.shape[1]], dtype=object)
labels[0:20, 0:20] = 'safe'
labels[12:16, 15:19] = 'goal1'
labels[15:19, 15:19] = 'goal2'
labels[9:13, 9:13] = 'goal3'
labels[0:4, 15:19] = 'goal4'

# override the labels
slp_med.labels = labels


from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# OpenAI Gym FrozenLake 8x8
# only the labelling function needs to be specified

# create a SlipperyGrid object
FrozenLake = SlipperyGrid(shape=[8, 8], initial_state=[0, 0], slip_probability=0.05, sink_states=[[7, 7]])

# defines the labelling image
labels = np.empty([FrozenLake.shape[0], FrozenLake.shape[1]], dtype=object)
labels[0:8, 0:8] = 'safe'
labels[2][3] = labels[3][5] = labels[4][3] = labels[5][1] = 'unsafe'
labels[5][2] = labels[5][6] = labels[6][1] = 'unsafe'
labels[6][4] = labels[6][6] = labels[7][3] = 'unsafe'
labels[0][7] = 'goal1'
labels[7][7] = 'goal2'

# override the labels
FrozenLake.labels = labels


FrozenLake.action_space = [
    "right",
    "up",
    "left",
    "down",
]




import os
import random
import numpy as np
from lcrl.environments.SlipperyGrid import SlipperyGrid
from lcrl.automata.ldba import LDBA
from lcrl.core.lcrl_core import LCRL
from lcrl.animator.animator import animate
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib import colors
from datetime import datetime


def train(
        MDP,
        LDBA,
        algorithm='ql',
        episode_num=2500,
        iteration_num_max=4000,
        discount_factor=0.95,
        learning_rate=0.9,
        nfq_replay_buffer_size=50,
        ddpg_replay_buffer_size=100000,
        decaying_learning_rate=False,
        epsilon=0.1,
        save_dir='./results',
        test=True,
        average_window=-1,
):
    learning_task = LCRL(MDP, LDBA, discount_factor, learning_rate, decaying_learning_rate, epsilon)

    if algorithm == 'ql':
        learning_task.train_ql(episode_num, iteration_num_max)
        import dill
        from lcrl.environments.MarsRoverDA import MarsRover
    elif algorithm == 'nfq':
        learning_task.train_nfq(episode_num, iteration_num_max, nfq_replay_buffer_size)
        import dill
        from lcrl.environments.MarsRoverDA import MarsRover
    elif algorithm == 'ddpg':
        learning_task.train_ddpg(episode_num, iteration_num_max, ddpg_replay_buffer_size)
        import dill
        import tensorflow as tf
        tf.get_logger().setLevel('ERROR')
        from lcrl.environments.MarsRoverCA import MarsRover
    else:
        raise NotImplementedError('New learning algorithms will be added to LCRL soon. The selected algorithm is not '
                                  'implemented yet.')

    if average_window == -1:
        average_window = int(0.03 * episode_num)

    plt.plot(learning_task.q_at_initial_state, c="royalblue")
    plt.xlabel('Episode Number')
    plt.ylabel('Value Function at The Initial State')
    plt.grid(True)
    if average_window > 0:
        avg = np.convolve(learning_task.q_at_initial_state, np.ones((average_window,)) / average_window, mode='valid')
        plt.plot(avg, c='darkblue')

    # saving the results
    results_path = os.path.join(os.getcwd(), save_dir[2:])
    dt_string = datetime.now().strftime("%d.%m.%Y_%H.%M.%S")
    results_sub_path = os.path.join(os.getcwd(), save_dir[2:], dt_string)
    if not os.path.exists(results_path):
        os.mkdir(results_path)
    os.mkdir(results_sub_path)
    plt.savefig(os.path.join(results_sub_path, 'convergence.png'))

    plt.show()

    if test:
        print('testing...')
        number_of_tests = 100
        learning_task.successes_in_test = 0
        for tt in range(number_of_tests):
            MDP.reset()
            LDBA.reset()
            # check if MDP current_state is a list or ndarray:
            if type(MDP.current_state) == np.ndarray:
                ndarray = True
                test_path = [MDP.current_state.tolist()]
            else:
                ndarray = False
                test_path = [MDP.current_state]
            iteration_num = 0
            while LDBA.accepting_frontier_set and iteration_num < iteration_num_max \
                    and LDBA.automaton_state != -1:
                iteration_num += 1
                if ndarray:
                    if algorithm == "nfq":
                        current_state = MDP.current_state.tolist() + [LDBA.automaton_state]
                    if algorithm == "ddpg":
                        prev_state = MDP.current_state
                        current_state = MDP.current_state.tolist() + [LDBA.automaton_state]
                else:
                    current_state = MDP.current_state + [LDBA.automaton_state]

                if learning_task.epsilon_transitions_exists:
                    product_MDP_action_space = learning_task.action_space_augmentation()
                else:
                    product_MDP_action_space = MDP.action_space

                if not algorithm == "ddpg":
                    Qs = []
                    if (not ndarray) and (str(current_state) in learning_task.Q.keys()):
                        for action_index in range(len(product_MDP_action_space)):
                            Qs.append(learning_task.Q[str(current_state)][product_MDP_action_space[action_index]])
                    elif ndarray:
                        for action_index in range(len(product_MDP_action_space)):
                            Qs.append(learning_task.Q[current_state[-1]].predict(
                                [MDP.current_state.tolist() + [action_index]]))
                    else:
                        Qs.append(0)
                    maxQ_action_index = random.choice(np.where(Qs == np.max(Qs))[0])
                    maxQ_action = product_MDP_action_space[maxQ_action_index]
                    # check if an epsilon-transition is taken
                    if learning_task.epsilon_transitions_exists and \
                            maxQ_action_index > len(learning_task.MDP.action_space) - 1:
                        epsilon_transition_taken = True
                    else:
                        epsilon_transition_taken = False
                    if epsilon_transition_taken:
                        next_MDP_state = learning_task.MDP.current_state if not ndarray else learning_task.MDP.current_state.tolist()
                        next_automaton_state = learning_task.LDBA.step(maxQ_action)
                    else:
                        next_MDP_state = learning_task.MDP.step(maxQ_action)
                        next_automaton_state = learning_task.LDBA.step(learning_task.MDP.state_label(next_MDP_state))
                        if ndarray:
                            next_MDP_state = next_MDP_state.tolist()
                else:
                    # action space bounds
                    lower_bound = -1.0
                    upper_bound = 1.0
                    tf_prev_state = tf.expand_dims(tf.convert_to_tensor(prev_state), 0)
                    sampled_actions = tf.squeeze(learning_task.Q[current_state[-1]](tf_prev_state))
                    sampled_actions = sampled_actions.numpy()
                    legal_action = np.clip(sampled_actions, lower_bound, upper_bound)
                    action = [np.squeeze(legal_action)]
                    if learning_task.epsilon_transitions_exists and \
                            LDBA.automaton_state in LDBA.epsilon_transitions.keys() and \
                            random.random() > 0.5:
                        epsilon_action = random.choice(product_MDP_action_space[2:])
                        action = [np.squeeze(
                            int(epsilon_action[-1]) + learning_task.upper_bound
                        )]
                        epsilon_transition_taken = True
                    else:
                        epsilon_transition_taken = False
                    # product MDP modification (for more details refer to https://bit.ly/LCRL_CDC_2019)
                    if epsilon_transition_taken:
                        next_automaton_state = LDBA.step(epsilon_action)
                        next_MDP_state = prev_state
                    else:
                        state = MDP.step(action)
                        next_MDP_state = state.tolist()
                        next_automaton_state = LDBA.step(MDP.state_label(next_MDP_state))

                    # product MDP: synchronise the automaton with MDP
                    prev_state = state.copy()

                test_path.append(next_MDP_state)
                if not epsilon_transition_taken:
                    LDBA.accepting_frontier_function(next_automaton_state)

                if not LDBA.accepting_frontier_set:
                    learning_task.successes_in_test += 1

        print('success rate in testing: ' + str(100 * learning_task.successes_in_test / number_of_tests) + '%')

    if isinstance(MDP, SlipperyGrid) and test:
        # plt.plot(learning_task.path_length, c='royalblue')
        # plt.xlabel('Episode Number')
        # plt.ylabel('Agent Traversed Distance from The Initial State')
        # plt.grid(True)
        # if average_window > 0:
        #     avg = np.convolve(learning_task.path_length, np.ones((average_window,)) / average_window, mode='valid')
        #     plt.plot(avg, c='darkblue')
        # plt.savefig(os.path.join(results_sub_path, 'traversed distance in the grid.png'))
        # plt.show()

        distinct_labels = np.unique(learning_task.MDP.labels)
        labels_dic = {}
        label_indx = 0
        bounds = [-0.9]
        cmap = plt.get_cmap('gist_rainbow')
        for label in distinct_labels:
            labels_dic[label] = label_indx
            bounds.append(bounds[-1] + 1)
            label_indx += 1
        color_map = cmap(np.linspace(0, 1, len(distinct_labels)))
        cmap = colors.ListedColormap(color_map)
        norm = colors.BoundaryNorm(bounds, cmap.N)
        labels_value = np.zeros([learning_task.MDP.shape[0], learning_task.MDP.shape[1]])
        for i in range(learning_task.MDP.shape[0]):
            for j in range(learning_task.MDP.shape[1]):
                labels_value[i][j] = labels_dic[learning_task.MDP.state_label([i, j])]
        patches = [mpatches.Patch(color=color_map[i], label=list(distinct_labels)[i]) for i in
                   range(len(distinct_labels))]
        plt.imshow(labels_value, interpolation='nearest', cmap=cmap, norm=norm)
        plt.legend(handles=patches, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        path_x, path_y = np.array(test_path).T
        plt.scatter(path_y, path_x, c='lime', edgecolors='teal')
        plt.scatter(path_y[0], path_x[0], c='red', edgecolors='black')
        plt.annotate('s_0', (path_y[0], path_x[0]), fontsize=15, xytext=(20, 20), textcoords="offset points",
                     va="center", ha="left",
                     bbox=dict(boxstyle="round", fc="w"),
                     arrowprops=dict(arrowstyle="->"))
        plt.title('This policy is synthesised by the trained agent')
        plt.savefig(
            os.path.join(results_sub_path, 'tested_policy.png'), bbox_inches="tight")
        plt.show()
        is_gif = input(
            'Would you like to create a gif for the the control policy? '
            'If so, type in "y", otherwise, type in "n". ')
        if is_gif == 'y' or is_gif == 'Y':
            animate(learning_task.MDP, test_path, results_sub_path, labels_value, cmap, norm, patches)
        print('\n---------------------------------\n')
        print('The results have been saved here:\n')
        print(results_sub_path)
        return learning_task

    if isinstance(MDP, MarsRover) and test:
        plt.imshow(MDP.background)
        path_x, path_y = np.array(test_path).T
        plt.scatter(path_y, path_x, c='lime', edgecolors='teal')
        plt.scatter(path_y[0], path_x[0], c='red', edgecolors='black')
        plt.annotate('s_0', (path_y[0], path_x[0]), fontsize=15, xytext=(20, 20), textcoords="offset points",
                     va="center", ha="left",
                     bbox=dict(boxstyle="round", fc="w"),
                     arrowprops=dict(arrowstyle="->"))
        plt.title('This policy is synthesised by the trained agent')
        plt.savefig(
            os.path.join(results_sub_path, 'tested_policy.png'), bbox_inches="tight")
        plt.show()
        # is_gif = input(
        #     'Would you like to create a gif for the the control policy? '
        #     'If so, type in "y", otherwise, type in "n". ')
        # if is_gif == 'y' or is_gif == 'Y':
        #     animate(learning_task.MDP, test_path, results_sub_path, labels_value, cmap, norm, patches)
        print('\n---------------------------------\n')
        print('The results have been saved here:\n')
        print(results_sub_path)
        return learning_task

    if algorithm == 'ql':
        with open(os.path.join(results_sub_path, 'learned_model.pkl'), 'wb') as learning_file:
            dill.dump(learning_task, learning_file)
        if test:
            with open(os.path.join(results_sub_path, 'test_results.pkl'), 'wb') as test_file:
                dill.dump(test_path, test_file)
        print('In order to load the learning results use the following command in Python console:')
        print('import dill')
        print("learned_model = dill.load(open('" + os.path.join(results_sub_path, 'learned_model.pkl') + "', 'rb'))")
        if test:
            print("tested_trace = dill.load(open('" + os.path.join(results_sub_path, 'test_results.pkl') + "', 'rb'))")
        print('\n---------------------------------\n')
        if learning_task.early_interruption == 0:
            print("Training finished successfully!")
        else:
            print("Training results have been saved successfully! [Note: training was interrupted by user]")
        return learning_task
    # TODO: change the save method and add nfq & ddpg
    return learning_task



from lcrl.environments.SlipperyGrid import SlipperyGrid
import numpy as np

# A modified version of OpenAI Gym FrozenLake
# only the labelling function needs to be specified
sinks = []
for i in range(12, 16):
    for j in range(15, 19):
        sinks.append([i, j])

# create a SlipperyGrid object
FrozenLake = SlipperyGrid(shape=[20, 20],
                          initial_state=[0, 10],
                          slip_probability=0.1,
                          sink_states=sinks
                          )

# define the labellings
labels = np.empty([FrozenLake.shape[0], FrozenLake.shape[1]], dtype=object)
labels[0:20, 0:20] = 'safe'
labels[4:8, 9:13] = 'unsafe'
labels[12:16, 15:19] = 'goal1'
labels[15:19, 15:19] = 'goal2'
labels[9:13, 9:13] = 'goal3'
labels[0:4, 15:19] = 'goal4'

# override the labels
FrozenLake.labels = labels

# FrozenLake doesn't have the action "stay"
FrozenLake.action_space = [
    "right",
    "up",
    "left",
    "down",
]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='lcrl',
    version='0.0.9.2',
    author="Hosein Hasanbeig",
    author_email="hosein.hasanbeig@icloud.com",
    keywords='rl, logic, environment, agent',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/grockious/lcrl',
    description='Logically-Constrained Reinforcement Learning',
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'tensorflow==2.7.0',
        'numpy',
        'matplotlib',
        'dill>=0.3.2',
        'imageio',
        'tqdm'
    ]
)



# Copyright (C) 2020 University of Oxford. All Rights Reserved.
# __all__ = ['ldba',
#            'minecraft_1',
#            'minecraft_2',
#            'minecraft_3',
#            'minecraft_4',
#            'minecraft_5',
#            'minecraft_6',
#            'minecraft_7',
#            'mars_rover_1_and_3',
#            'mars_rover_2_and_4',
#            'surveillance',
#            'slp_easy',
#            'slp_hard',
#            'frozenlake_reach_avoid',
#            'frozenlake_sequential',
#            'pacman_foods'
#            ]



# Copyright (C) 2020 University of Oxford. All Rights Reserved.
# __all__ = ['minecraft',
#            'mars_rover_discrete_action',
#            'mars_rover_continuous_action',
#            'robot_surv',
#            'slp_sml',
#            'slp_med',
#            'slp_lrg',
#            'frozenlake_1',
#            'frozenlake_2',
#            'frozenlake_3',
#            'frozenlake_4',
#            'frozenlake_5',
#            'frozenlake_6',
#            'pacman'
#            ]
